import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype
from scipy.stats import norm

from .IRR_result import IRR_result


def kappa2(ratings, weight, sort_levels=False):
    """Cohen’s Kappa and weighted Kappa for two raters

    Parameters
    ----------
    ratings: array_like
        subjects * raters array or dataframe
    weight: {"unweighted", "equal", "squared"}
        either a character string specifying one predefined set of weights or a numeric vector with own weights.
    sort_levels: bool

    """
    ratings = pd.DataFrame(ratings)  # make sure ratings is a DataFrame

    ratings.dropna(inplace=True)

    ns = ratings.shape[0]
    nr = ratings.shape[1]

    if nr > 2:
        raise Exception("Number of raters exceeds 2. Try kappam_fleiss or kappam_light")

    r1, r2 = ratings.iloc[:, 0], ratings.iloc[:, 1]

    if is_numeric_dtype(r1) or is_numeric_dtype(r2):
        sort_levels = True

    levels = np.unique(ratings)

    if sort_levels:
        levels = np.sort(levels)

    r1 = pd.Categorical(r1, categories=levels)
    r2 = pd.Categorical(r2, categories=levels)

    ttab = pd.crosstab(r1, r2, dropna=False).values
    nc = ttab.shape[1]

    if not isinstance(weight, str):
        w = 1 - (np.asarray(weight) - min(weight)) / (max(weight) - min(weight))
    elif weight == "equal":
        w = np.arange((nc-1), -1, -1) / (nc-1)
    elif weight == "squared":
        w = 1 - np.arange(0, nc, 1)**2 / (nc-1)**2
    elif weight == "unweighted":
        w = np.zeros(nc)
        w[0] = 1

    nw = len(w)
    wvec = np.append(np.sort(w), w[1:])
    weight_tab = np.zeros((nw, nw))

    for i in range(nw):
        weight_tab[i, :] = wvec[(nw - i - 1):(2 * nw - i - 1)]

    agreeP = np.sum(ttab * weight_tab) / ns

    tm1 = np.sum(ttab, 1)
    tm2 = np.sum(ttab, 0)

    eij = np.outer(tm1, tm2) / ns
    chanceP = np.sum(eij * weight_tab) / ns

    # Kappa for 2 raters
    value = (agreeP - chanceP) / (1 - chanceP)

    # Compute statistics
    wi = np.sum(np.broadcast_to(tm2/ns, (nc, nc)).T * weight_tab, 0)
    wj = np.sum(np.repeat(tm1/ns, nc).reshape((nc, nc)).T * weight_tab, 1)

    var_matrix = (eij / ns) * (weight_tab - np.sum(np.meshgrid(wi, wj), axis=0).T) ** 2

    var_kappa = (np.sum(var_matrix) - chanceP ** 2) / (ns * (1 - chanceP) ** 2)

    SE_kappa = np.sqrt(var_kappa)
    u = value / SE_kappa

    pvalue = 2 * (1 - norm.cdf(abs(u)))

    # return kappa2_result(ns, nr, value, u, p_value, weight)
    method = f"Cohen's Kappa for 2 Raters (Weights: {weight})"
    return IRR_result(method, ns, nr, "Kappa", value, u, "z", pvalue)



import numpy as np
import pandas as pd
from scipy.stats import chi2

from .IRR_result import IRR_result


def bhapkar(ratings):
    """Calculate the percentage agreement among raters

    Parameters
    ----------
    ratings: array_like
        subjects * raters array or dataframe

    """
    ratings = pd.DataFrame(ratings)  # make sure ratings is a DataFrame

    ratings.dropna(axis=0, inplace=True)

    ns = ratings.shape[0]
    nr = ratings.shape[1]

    if nr > 2:
        raise Exception("Number of raters exceeds 2. Try kappam_fleiss or kappam_light")

    r1, r2 = ratings.iloc[:, 0], ratings.iloc[:, 1]

    # find factor levels
    levels = np.unique(ratings)

    r1 = pd.Categorical(r1, categories=levels)
    r2 = pd.Categorical(r2, categories=levels)

    # compute table
    ttab = pd.crosstab(r1, r2, dropna=False)

    # get marginals
    row_sums = ttab.sum(axis=1)[:-1]
    col_sums = ttab.sum(axis=0)[:-1]

    # compute d matrix
    dmat = np.tile(row_sums - col_sums, [len(row_sums), 1])

    # setup delta matrix
    delta = np.zeros((len(row_sums), len(row_sums)))
    np.fill_diagonal(delta, row_sums + col_sums)

    # dump last category from smx table
    smx = ttab.iloc[:-1, :-1]

    # compute w matrix
    w = delta - smx - smx.T - (dmat * dmat.T) / ns
    w1 = np.linalg.inv(w)

    # compute chisq-value
    chimat = dmat * dmat.T * w1

    # test statistics
    x_value = np.sum(chimat)
    df1 = len(ttab) - 1
    pvalue = 1 - chi2.cdf(x_value, df1)

    method = "Bhapkar marginal homogeneity"
    stat_name = f"Chisq({df1})"
    return IRR_result(method, ns, nr, "Chisq", x_value, x_value, stat_name, pvalue)



import numpy as np
import pandas as pd

from .IRR_result import IRR_result


def iota(ratings, scale_data, standardize=False):
    """Computes iota as an index of interrater agreement of quantitative or nominal multivariate observations.

    Parameters
    ----------
    ratings: list
        list with subjects * raters array or dataframe
    scale_data: {"quantitative", "nominal"}
        a character string specifying if the data is '"quantitative"' (default) or '"nominal"'.
    standardize: bool
       a logical indicating whether quantitative data should be z-standardized within each variable before the
       computation of iota.

    """
    detail = None
    for i, rating in enumerate(ratings):
        rating = pd.DataFrame(rating)
        ratings[i] = rating.dropna()

    ns = ratings[0].shape[0]
    nr = ratings[0].shape[1]

    nvar = len(ratings)  # number of variables

    if scale_data == "quantitative":
        if standardize:
            for i, rating in enumerate(ratings):
                x = np.ravel(rating)
                ratings[i] = (rating - x.mean()) / x.std()
            detail = "Variables have been z-standardized before the computation"
        ratinglist = [np.array(rating) for rating in ratings]  # Take original as new rating-structure
    elif scale_data == "nominal":
        ratinglist = []
        dummyn = 0
        for i, rating in enumerate(ratings):
            # How many levels were used?
            levels = np.unique(rating.values)

            for level in levels:
                # Build new rating-structure
                ratinglist.append(np.zeros((ns, nr)))
                ratinglist[dummyn][ratings[i] == level] = 1 / np.sqrt(2)
                dummyn += 1
    else:
        raise Exception("Please choose quantitative or nominal.")

    # Compute coefficient
    doSS, deSS = 0, 0

    for rating in ratinglist:
        SSt = np.cov(rating.ravel()) * (ns * nr - 1)
        SSw = np.sum(np.apply_along_axis(np.cov, 1, rating) / ns) * ns * (nr - 1)
        SSc = np.cov(np.mean(rating, 0)) * ns * (nr - 1)

        doSS = doSS + SSw
        deSS = deSS + ((nr - 1) * SSt + SSc)

    coeff = 1 - (nr * doSS) / deSS

    return IRR_result(f"iota for {scale_data} ({nvar} variable(s))", ns, nr, "iota", coeff, detail=detail)



import numpy as np
import pandas as pd
from scipy.special import comb
from scipy.stats import norm

from .kappa2 import kappa2
from .IRR_result import IRR_result


def kappam_light(ratings):
    """Computes Light's Kappa as an index of interrater agreement between m raters on categorical data.

    Parameters
    ----------
    ratings: array_like
        subjects * raters array or dataframe

   """
    ratings = pd.DataFrame(ratings)
    ns = ratings.shape[0]
    nr = ratings.shape[1]

    ratings.dropna(inplace=True)  # drop nans

    kappas = []
    for i in range(nr - 1):
        for j in range(i + 1, nr):
            kappas.append(kappa2(ratings.iloc[:, [i, j]], weight="unweighted").value)

    value = np.mean(kappas)

    # Variance & Computation of p-value
    lev = np.unique(ratings)
    levlen = len(lev)

    dis = []
    disrater = []
    for nri in range(nr - 1):
        for nrj in range(nri + 1, nr):
            for i in range(levlen):
                for j in range(levlen):
                    if i != j:
                        r1i = np.sum(ratings.iloc[:, nri] == lev[i])
                        r2j = np.sum(ratings.iloc[:, nrj] == lev[j])
                        dis.append(r1i * r2j)
            disrater.append(np.sum(dis))
            dis = []  # reset dis

    prod = 1
    for i in disrater:
        prod *= int(i)  # numpy overflows here so we need a Python int
    B = len(disrater) * prod

    chanceP = 1 - B / ns**(comb(nr, 2) * 2)
    varkappa = chanceP / (ns * (1 - chanceP))

    SEkappa = np.sqrt(varkappa)
    u = value / SEkappa
    pvalue = 2 * (1 - norm.cdf(np.abs(u)))

    method = "Light's Kappa for m Raters"
    return IRR_result(method, ns, nr, "Kappa", value, u, "z", pvalue)



import numpy as np
import pandas as pd
from scipy.stats import chi2

from .IRR_result import IRR_result


def rater_bias(ratings):
    """Calculates a coefficient of systematic bias between two raters.

    Parameters
    ----------
    ratings: array_like
        n x 2 matrix of classification scores into c categories.

    """
    ratings = pd.DataFrame(ratings)

    ns = ratings.shape[0]
    nr = ratings.shape[1]

    if nr > 2:
        raise Exception("More than two raters, cannot compute")

    rbx = pd.crosstab(ratings.iloc[:, 0], ratings.iloc[:, 1])
    rbb = np.sum(np.triu(rbx, k=len(rbx)//2 - 1))
    rbc = np.sum(np.tril(rbx, k=-len(rbx)//2 + 1))
    rb = np.abs(rbb/(rbb+rbc))
    rbstat = (rbb - rbc)**2 / (rbb + rbc)

    pvalue = 1 - chi2.cdf(rbstat, 1)

    return IRR_result("Rater bias coefficient", ns, nr, "Ratio", rb, rbstat, f"Chisq(1)", pvalue)



from pprint import pformat

import numpy as np
import pandas as pd
import statsmodels.api as sm
from scipy.stats import f
from statsmodels.formula.api import ols

from .IRR_result import IRR_result


def rel_inter_intra(ratings, nraters, rho_inter=0., rho_intra=0.8, conf_level=0.95):
    """Calculates inter- and intra-rater reliability coefficients.

    Parameters
    ----------
    ratings: array_like
        dataframe or array of rater by object scores with consecutive measurements for each rater in adjacent columns
    nraters: int
        number of raters
    rho_inter: float
        null hypothesis value for the inter-rater reliability coefficient
    rho_intra: float
        null hypothesis value for the intra-rater reliability coefficient
    conf_level: float
        confidence level for the one-sided confidence interval reported

    """
    ratings = pd.DataFrame(ratings)

    ns = ratings.shape[0]
    nmeas = ratings.shape[1] / nraters

    data = np.vstack([np.tile(np.arange(1, ns + 1), int(nraters * nmeas)),
                      np.repeat(np.arange(1, nraters + 1), ns * nmeas),
                      np.tile(np.repeat(np.arange(1, nmeas + 1), ns), nraters),
                      np.ravel(ratings, order="F")]).T
    frame1 = pd.DataFrame(data, columns=["Subject", "Rater", "Repetition", "Result"])
    frame1["Subject"] = frame1["Subject"].astype("category")
    frame1["Rater"] = frame1["Rater"].astype("category")
    frame1["Repetition"] = frame1["Repetition"].astype("category")

    nn = ns  # aliases for compatibility with Eliasziw et al. notation
    tt = nraters
    mm = nmeas

    mean_squares = []
    aov = ols('Result ~ Subject * Rater', data=frame1).fit()
    aov_table = sm.stats.anova_lm(aov, typ=2)
    mean_squares.append(aov_table["sum_sq"] / aov_table["df"])

    for rater_act in frame1["Rater"].cat.categories:
        aov_act = ols('Result ~ Subject', data=frame1[frame1["Rater"] == rater_act]).fit()
        aov_act_table = sm.stats.anova_lm(aov_act, typ=2)
        mean_squares.append((aov_act_table["sum_sq"] / aov_act_table["df"]).iloc[-1])

    MSS = mean_squares[0][0]
    MSR = mean_squares[0][1]
    MSSR = mean_squares[0][2]
    MSE = mean_squares[0][3]

    # the same for random and fixed, see table 2 (p. 780) and 3 (p.281)
    MSEpart = np.array(mean_squares[1:])
    sighat2Srandom = (MSS - MSSR) / (mm * tt)
    sighat2Rrandom = (MSR - MSSR) / (mm * nn)
    sighat2SRrandom = (MSSR - MSE) / mm

    # the same for random and fixed, see table 2 (p. 780) and 3 (p. 281)
    sighat2e = MSE
    sighat2Sfixed = (MSS - MSE) / (mm * tt)
    sighat2SRfixed = (MSSR - MSE) / mm

    # the same for random and fixed, see table 2 (p. 780) and 3 (p. 281)
    sighat2e_part = MSEpart
    rhohat_inter_random = sighat2Srandom / (sighat2Srandom + sighat2Rrandom + sighat2SRrandom + sighat2e)
    rhohat_inter_fixed = (sighat2Sfixed - sighat2SRfixed / tt) / (sighat2Sfixed + (tt - 1) *
                                                                  sighat2SRfixed / tt + sighat2e)
    rhohat_intra_random = (sighat2Srandom + sighat2Rrandom + sighat2SRrandom) / (sighat2Srandom + sighat2Rrandom +
                                                                                 sighat2SRrandom + sighat2e)
    rhohat_intra_fixed = (sighat2Sfixed + (tt - 1) * sighat2SRfixed / tt) / (sighat2Sfixed + (tt - 1) *
                                                                             sighat2SRfixed / tt + sighat2e)
    rhohat_intra_random_part = (sighat2Srandom + sighat2Rrandom + sighat2SRrandom) / (sighat2Srandom + sighat2Rrandom +
                                                                                      sighat2SRrandom + sighat2e_part)
    rhohat_intra_fixed_part = (sighat2Sfixed + (tt - 1) * sighat2SRfixed / tt) / (sighat2Sfixed + (tt - 1) *
                                                                                  sighat2SRfixed / tt + sighat2e_part)

    F_inter = (1 - rho_inter) * MSS / ((1 + (tt - 1) * rho_inter) * MSSR)
    F_inter_p = 1 - f.cdf(F_inter, nn - 1, (nn - 1) * (tt - 1))
    alpha = 1 - conf_level

    nu1 = (nn - 1) * (tt - 1) * (tt * rhohat_inter_random * (MSR - MSSR) + nn * (1 + (tt - 1) * rhohat_inter_random) *
          MSSR + nn * tt * (mm - 1) * rhohat_inter_random * MSE) ** 2 / ((nn - 1) * (tt * rhohat_inter_random) ** 2 *
          MSR ** 2 + (nn * (1 + (tt - 1) * rhohat_inter_random) - tt * rhohat_inter_random) ** 2 * MSSR ** 2 +
          (nn - 1) * (tt - 1) * (nn * tt * (mm - 1)) * rhohat_inter_random ** 2 * MSE ** 2)
    nu2 = (nn - 1) * (tt - 1) * (nn * (1 + (tt - 1) * rhohat_inter_fixed) * MSSR + nn * tt * (mm - 1) *
           rhohat_inter_fixed * MSE) ** 2 / ((nn * (1 + (tt - 1) * rhohat_inter_fixed)) ** 2 * MSSR ** 2 + (nn - 1) *
          (tt - 1) * (nn * tt * (mm - 1)) * rhohat_inter_fixed ** 2 * MSE ** 2)

    F1 = f.ppf(1 - alpha, nn - 1, nu1)
    F2 = f.ppf(1 - alpha, nn - 1, nu2)

    lowinter_random = nn * (MSS - F1 * MSSR) / (nn * MSS + F1 * (tt * (MSR - MSSR) + nn * (tt - 1) *
                                                                 MSSR + nn * tt * (mm - 1) * MSE))
    lowinter_random = min(lowinter_random, 1)

    lowinter_fixed = nn * (MSS - F2 * MSSR) / (nn * MSS + F2 * (nn * (tt - 1) * MSSR + nn * tt * (mm - 1) * MSE))
    lowinter_fixed = min(lowinter_fixed, 1)

    F_intra = (1 - rho_intra) * MSS / ((1 + (mm - 1) * rho_intra) * MSE * tt)
    F_intra_p = 1 - f.cdf(F_intra, nn - 1, nn * (mm - 1))

    F_intra_part = (1 - rho_intra) * MSS / ((1 + (mm - 1) * rho_intra) * sighat2e_part * tt)
    F_intra_part_p = 1 - f.cdf(F_intra_part, nn - 1, nn * (mm - 1))

    F3 = 1 - f.ppf(1 - alpha, nn - 1, nn * (mm - 1))

    low_intra = (MSS / tt - F3 * MSE) / (MSS / tt + F3 * (mm - 1) * MSE)
    low_intra = min(low_intra, 1)

    F4 = f.ppf(1 - alpha, nn - 1, nn * (mm - 1))
    low_intra_part = (MSS / tt - F4 * MSEpart) / (MSS / tt + F4 * (mm - 1) * MSEpart)

    low_intra_part = np.clip(low_intra_part, 1, None)

    SEMintra = np.sqrt(MSE)
    SEMintra_part = np.sqrt(MSEpart)
    SEMinter_random = np.sqrt(sighat2Rrandom + sighat2SRrandom + sighat2e)
    SEMinter_fixed = np.sqrt(sighat2SRfixed + sighat2e)

    detail = {"rohat": {"rhohat_inter_random": rhohat_inter_random, "rhohat_intra_random": rhohat_intra_random,
                        "rhohat_inter_fixed": rhohat_inter_fixed, "rhohat_intra_fixed": rhohat_intra_fixed,
                        "rhohat_intra_random_part": rhohat_intra_random_part,
                        "rhohat_intra_fixed_part": rhohat_intra_fixed_part},
              "Fs": {"F_inter": F_inter, "F_intra": F_intra, "F_intra_part": F_intra_part},
              "pvalue": {"F_inter_p": F_inter_p, "F_intra_p": F_intra_p, "F_intra_part_p": F_intra_part_p},
              "lowvalue": {"lowinter_random": lowinter_random, "lowinter_fixed": lowinter_fixed,
                           "low_intra": low_intra, "low_intra_part": low_intra_part},
              "sem ": {"SEMintra": SEMintra, "SEMintra_part": SEMintra_part, "SEMinter_random": SEMinter_random,
                       "SEMinter_fixed": SEMinter_fixed}}

    for key, value in detail.items():  # do some rounding in the nested dicts
        detail[key] = {nkey: np.round(nvalue, 4) for nkey, nvalue in value.items()}

    return IRR_result("Inter/Intrarater reliability", ns, nraters, "rhohat", np.NaN, detail=pformat(detail))



from dataclasses import dataclass, asdict
from typing import Any


@dataclass
class IRR_result:
    method: str
    subjects: int
    raters: int
    irr_name: str
    value: float
    statistic: float = None
    stat_name: str = None
    pvalue: float = None
    detail: Any = None
    error: str = None

    def to_dict(self):
        return asdict(self)

    def __repr__(self):
        model_string = "=" * 50 + "\n"
        model_string += f"{self.method}".center(50, " ") + "\n"
        model_string += "=" * 50 + "\n"
        model_string += f"Subjects = {self.subjects}\n  Raters = {self.raters}\n"
        model_string += f"{self.irr_name:>8}" + f" = {self.value:.3f}\n\n"

        if self.statistic is not None:
            model_string += f"{self.stat_name:>8} = {self.statistic:.3f}\n"
            model_string += f" p-value = {self.pvalue:.3f}\n"

        if self.detail is not None:
            model_string += f"\n{self.detail}\n"

        if self.error is not None:
            model_string += f"\n{self.error}\n"
        model_string += "=" * 50 + "\n"
        return model_string



import os
from setuptools import setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
	name             		= 'pyirr',
	version          		= '0.84.1.1',
	description      		= 'Python implementation of the R package IRR',
	author           		= 'Rick de Klerk',
	author_email     		= 'r.de.klerk@umcg.nl',
	url              		= 'https://gitlab.com/Rickdkk/pyrr',
	download_url     		= 'https://gitlab.com/Rickdkk/pyrr',
	packages         		= ['pyirr'],
	package_data     		= {"pyirr": ['pyirr/data/*']},
	include_package_data 	= True,
	long_description 		= read("README.rst"),
	license 				= 'GNU GPLv3',
	keywords         		= ['statistics'],
	classifiers      		= [],
	install_requires 		= ["numpy", "scipy", "pandas", "statsmodels"]
)
